#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

###
#COPYRIGHT © 2013 THE OHIO STATE UNIVERSITY
#ALL RIGHTS RESERVED
#
#PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
#REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
#EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
#LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
#DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
#THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
#PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
#SPECIFIC, WRITTEN PRIOR AUTHORIZATION.
#
#THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
#STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
#WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
#IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
#UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
#INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
#CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
#EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGES.
###


import rpy2 as rpy2
import rpy2.robjects as robjects
import sys
import marshal
import os 
import gzip
import string
import matplotlib
import numpy
matplotlib.use('Agg')
import matplotlib.pyplot as plt

class MethMAGE():
    #=============================================
    # GLOBAL CLASS VARIABLES
    #=============================================

    # holds the index of methylation values 
    __sssnroi_ndx = None

    # holds the position of cpg sites
    __ssnroi_pos_ndx = None

    # holds the position of chr 
    __chr_ndx = None

    # holds the position of start 
    __start_ndx = None

    # holds the position of stop 
    __stop_ndx = None

    # holds the position of loci name 
    __loc_ndx = None

    # holds the position of refseq name 
    __refseq_ndx = None

    # holds the position of gene name 
    __gene_ndx = None

    # holds the minimum methylation among all samples 
    __both_grps_min = None

    # holds the number of loci used for fdr 
    __fdr_cnt = None

    # list of groupA file handles
    __ssnroisA = None

    # list of groupB file handles
    __ssnroisB = None

    # size of groupA
    __size_ssnroisA = None

    # size of groupB
    __size_ssnroisB = None



    ### output file variables
    __out = None
    __fltrd_out = None
    __plots_out = None 

    ### group names
    __Alabel = None
    __Blavel = None
    


    #=============================================
    # CLASS METHODS
    #=============================================

    def __init__(self):
        """
        Constructor. Instantiates the cg binner.
        """
        return

    #####
    ###
    #   GEE function
    ###
    #####
    def __GEE(self,mA,mB,nSize):
        """
        R based GEE function from geepack

        Returns a p-value for the null-hypothesis that the group average between the two matrices are equal.
        Assumes that rows are patients and columns are loci.
        
        possible returns of  (p-val, corrected-p-value):
        a- return (1,-1) if sum of both matrices is zero , if sum_siteVAL==0.0
        b- return (float(nlme[0]), float (fcval)) if test runs without errors
        c- return (None,None) if test fails
        d- return (None,None) if the whole function fails at any point
        
        
        """
        import sys   
        # import geepack library
        robjects.r('library(geepack)')
        
        # create R vector with group ID for every site
        numA = sum(map(len,mA))
        numB = sum(map(len,mB))
        groupID = '\''+'\',\''.join(['A']*numA+['B']*numB)+'\''
        robjects.r('AB <- c('+groupID+')')
        
        # create R vector with sample ID for every site
        sampID = []
        for i,l in enumerate(mA+mB):
            for j in l:
                sampID.append(i)
        sIDstr = '\''+'\',\''.join(map(str,sampID))+'\''
        robjects.r('SAMP <- c('+sIDstr+')')
        
        # create R vector with number of CpGs for each sample
        pos = []
        for vA in mA: pos.extend(xrange(len(vA)))
        for vB in mB: pos.extend(xrange(len(vB)))
        if len(pos) <= 0: return (None,None)
        sitePOS = ','.join(map(str,pos))
        robjects.r('POS <- c('+sitePOS+')')

        # create R vector with CpG position for each sample
        cgp = []
        for vA in mA: cgp.extend(vA)
        for vB in mB: cgp.extend(vB)
        if len(pos) <= 0: return (None,None)
        siteCGP = ','.join(map(str,cgp))
        robjects.r('CGP <- c('+siteCGP+')')

        # create R vector with site VAL for every site
        val = []
        for vA in mA: val.extend(vA)
        for vB in mB: val.extend(vB)
        if len(val) <= 0: return (None,None)
        siteVAL = ','.join(map(str,val))
        robjects.r('VAL <- c('+siteVAL+')')
        
        #transform val into binary: 0 if val == 0, 1 otherwise
        bval = []
        for i in val: 
            if float(i) != 0: bval.append(1)
            elif float(i) == 0: bval.append(0)
            else:
                print 'Weird value found in val \n',val 
        sitebVAL = ','.join(map(str,bval))
        robjects.r('bVAL <- c('+sitebVAL+')')

        sum_siteVAL = sum( map(float, siteVAL.split(',')) )        
        if sum_siteVAL==0.0: 
            return (1,-1)

        # create R data frame
        robjects.r('DATA = data.frame(group = AB, pos = POS, val = VAL, samp = SAMP, bval = bVAL)')
       

        # pipe R data frame through NLME model
        rslt_ls =[]
        try:
            ###
            # implement GEE
            ###
            gee = robjects.r(
            '''
            test = summary(geeglm(val ~ factor(group), id = factor(samp), data = DATA, corstr="ar1"  ))$coefficients[4][[1]][2]
            '''
            )
            ###
            # implement logit link test
            ###       
     
            ll = robjects.r(
            '''
            test = summary(geeglm(bval ~ factor(group), id = factor(samp), data = DATA, corstr="ar1", family=binomial  ))$coefficients[4][[1]][2]
            '''
            )
            

            pval = float(gee[0])
            llval = float(ll[0])
                
            robjects.r('p <-' + str(pval))
            robjects.r('nval <-' + str(nSize) )
            robjects.r('fcp <- p.adjust(p,"fdr", nval )')
            fcval_list = str(robjects.r('print(fcp)'))
            
            fcval = fcval_list.split(' ')[1]

            return (float(gee[0]), float (fcval),float(llval))
        except (rpy2.rinterface.RRuntimeError),err:
            return (None,None,None)

    #####
    ###
    #    Plot loci
    ###
    #####
    def __plotloci(self,cgpos, Bval, Aval, outname,title):
        plt.figure()
        if len(cgpos) == 0 or len(Bval) == 0 or len(Aval) == 0:
            return
        
	Agroup = []
        #plot each sample in groupA
        for samp in Aval:
            print 'length cgpos,Aval: ',len(cgpos),len(samp)
            print 'cgpos is: \n',map(float,cgpos)
            print 'samp is: \n',map(float,samp)
            plt.plot(map(float,cgpos),map(float,samp), '.r')

        Bgroup = []
        for samp in Bval:
            plt.plot(cgpos,samp, '.b')

        #plot average lines for each group
        Aavg = []
        Amax = []
        Amin = []
        Bmax = []
        Bmin = []
        for i in range(0,len(Aval[0])):
            loctmp = []
            for j,samp in enumerate(Aval):
                loctmp.append(float(Aval[j][i]))

            Aavg.append(numpy.mean(loctmp))
                
            Amax.append(max(loctmp))
            Amin.append(min(loctmp))

        Bavg = []
        for i in range(0,len(Bval[0])):
            loctmp = []
            for j,samp in enumerate(Bval):
                loctmp.append(float(Bval[j][i]))

            Bavg.append(numpy.mean(loctmp))

            Bmax.append(max(loctmp))
            Bmin.append(min(loctmp))

        for i,v in enumerate(Amax):
            plt.plot([cgpos[i],cgpos[i]],[Amin[i],Amax[i]],marker='None',alpha=.1,color='#EE82EE',lw=2)
            plt.plot([cgpos[i],cgpos[i]],[Bmin[i],Bmax[i]],marker='None',alpha=.1,color='#00FA9A',lw=2)
        
        #plot average
        Aplt, = plt.plot(cgpos,Aavg,marker='o',color='#EE82EE',markersize=5,markeredgecolor='k',mec='None',fillstyle='full',ls='dotted',lw=2)
        Bplt, = plt.plot(cgpos,Bavg,marker='s',color='#00FA9A',markersize=5,markeredgecolor='k',mec='None',fillstyle='full',ls='dotted',lw=2)
        Amavg = numpy.mean(Aavg)
        Bmavg = numpy.mean(Bavg)
        
	#add title and labels
        plt.title(title)
        plt.legend([Aplt, Bplt],[str(args.Alabel[0])+': AVG = '+str(Amavg), str(args.Blabel[0])+': AVG = '+str(Bmavg)])
        plt.xlabel('CpG position')
        plt.ylabel('Methylation Signal')
        
	#output image
        plt.savefig(outname,format='png')
        
        return

    #####
    ###
    #       MethMAGE main function
    ###
    #####

    def methmage(self,ssnroiAin,ssnroiBin,Alabel,Blabel,out,plots_out):


        #####
        ###
        # Load filehandles into group objects ssnroisA and 
    	# ssnroisB and process file obtain index locations,
    	# meth value minimums, etc
        ###
        #####
        grp_minA = 1

        #set global variables
        try:
            fltrd_out = out + '.fltrd'           
            self.__out = open(out,'w')        
            self.__fltrd_out = open(fltrd_out,'w')
    
        except (IOError,TypeError),err:
            print 'CAN NOT OPEN OUTPUT: ',out
            return
        
        plots_out = ''.join(plots_out)
	self.__ssnroisA = []
        self.__ssnroisB = []
        self.__itlistA = []
        self.__itlistB = []
        self.__Alabel = Alabel
        self.__Blavel = Blabel
        self.__size_ssnroisA = len(ssnroiAin)-1
        self.__size_ssnroisB = len(ssnroiBin)-1


        #hardcode data column name
        ssnroi_col_name='ssnroi' 
        ssnroi_pos_name='pos'
        chr_name='chr'
        start_name='start'
        stop_name='stop'
        loc_name='name'
        refseq_name='refseq'
        gene_name='genesym'
        nametest = True
        refseqtest = True
        genetest = True

        ###
        # Initial processing of the files
        ###

        ### Process groupA samples:
        print 'Loading samples from groupA into memory:'
        print 'Processing samples:'
        for i,A in enumerate(ssnroiAin):
            try: fh = gzip.open(A,'r')
            except (IOError),err: 
                print 'Read error for file ',A
                return
            print '\t',A

            file_list=[]
            # get data from all lines of each file
            for ln in fh :  
                line_flds = ln.strip('\n').split('\t')
                file_list.append(line_flds)  

                if (ln[0] == '#') : 
                    head_line = line_flds
                    line_flds = [ f.strip('#') for f in line_flds ]
                    self.__sssnroi_ndx = line_flds.index(ssnroi_col_name)
                    self.__ssnroi_pos_ndx = line_flds.index(ssnroi_pos_name)
                    self.__chr_ndx = line_flds.index(chr_name)
                    self.__start_ndx = line_flds.index(start_name)
                    self.__stop_ndx = line_flds.index(stop_name)
                    try:self.__loc_ndx = line_flds.index(loc_name)
                    except (ValueError): nametest = False
                    try:self.__refseq_ndx = line_flds.index(refseq_name)
                    except (ValueError): refseqtest = False
                    try:self.__gene_ndx = line_flds.index(gene_name)
                    except (ValueError): genetest = False   
                    continue            
                loci_ssnroi_vals = []
                non_zero_vals = []
                if ( string.find(line_flds[self.__sssnroi_ndx], 'NA') != -1) :
                    continue
                else : 
                    loci_ssnroi_vals = map (float, line_flds[self.__sssnroi_ndx].split(','))
                

                non_zero_vals = [x for x in sorted (loci_ssnroi_vals) if x!=0.0]
                if len (non_zero_vals) != 0 : 
                    ln_min = min (non_zero_vals)  
                    if ln_min < grp_minA : grp_minA = ln_min
            print 'Length of sample ',A,' is ',len(file_list)
           
            #seek to begining of fh
            fh.seek(0)
            #CREATE TWO ITERATORS
            #add normal iterator to ssnroisA list
            try:
                itobj = fh.__iter__()
                self.__ssnroisA.append(itobj)
                self.__itlistA.append(fh.__iter__())
            except (TypeError),err: 
                print 'ERROR: Problem loading data for sample ',A
                return

        # Process groupB samples
        print 'Loading samples from groupB into memory:'
        print 'Processing samples:'
        grp_minB = 1
        for i,B in enumerate(ssnroiBin):
            try: fh = gzip.open(B,'r')
            except (IOError),err: 
                print 'Read error for file ',B
                return
            print '\t',B

            file_list=[]
            for ln in fh :  
                line_flds = ln.strip('\n').split('\t')
                file_list.append ( line_flds )  

                if (ln[0] == '#') : 
                    head_line = line_flds
                    self.__sssnroi_ndx = line_flds.index(ssnroi_col_name)
                    self.__ssnroi_pos_ndx = line_flds.index(ssnroi_pos_name)
                    continue            
                loci_ssnroi_vals = []
                non_zero_vals = []
                if ( string.find(line_flds[self.__sssnroi_ndx], 'NA') != -1) :
                    continue
                else: 
                    loci_ssnroi_vals = map (float, line_flds[self.__sssnroi_ndx].split(','))
            
                non_zero_vals = [x for x in sorted (loci_ssnroi_vals) if x!=0.0]
                if len (non_zero_vals) != 0 : 
                    ln_min = min (non_zero_vals)  
                    if ln_min < grp_minB : grp_minB = ln_min
            

            print 'Length of sample ',B,' is ',len(file_list)    

            #seek to begining of fh
            fh.seek(0)
            #CREATE TWO ITERATORS
            #add normal iterator to ssnroisB list
            try: 
                itobj = fh.__iter__()
                self.__ssnroisB.append(itobj)
                self.__itlistB.append(fh.__iter__())
            except (TypeError),err:
                print 'ERROR: Problem loading data for sample ',B
                return

        #set min value from both groups
        if grp_minA < grp_minB: 
            self.__both_grps_min = grp_minA 
        else :
           self.__both_grps_min = grp_minB

        #set test vals for each group
        self.__testValA = self.__size_ssnroisA * self.__both_grps_min 
        self.__testValB = self.__size_ssnroisB * self.__both_grps_min 

        #set loc names
        #process name ndx
        if nametest:
            self.__loc_name_ndx = self.__loc_ndx
        if refseqtest:
            self.__loc_name_ndx = self.__refseq_ndx
        if genetest:
            self.__loc_name_ndx = self.__gene_ndx
        
        #output headers
        sub_header= ('\t'.join([x for x in head_line if x != ssnroi_col_name and x != ssnroi_pos_name ])).strip('\n')
        header =  sub_header +'\tssnlme-p\tfdr_pval\tClass\tnonzero_frac\tavgGrpA\tavgGrpB\n'
        self.__out.write( header)
        self.__fltrd_out.write( header)

  



        ###
        # find loci that can support GEE
        ###
        
	# use first file in A to iterate over 
        invalidComparisons = 0
        allComparisons = 0
        for itline in self.__ssnroisA[0]:
            #set empty lists
            A = []
            B = []
            allComparisons += 1
            #add iterator line to A list
            A.append(itline.strip('\n').split('\t'))
            for i,fhA in enumerate(self.__ssnroisA):
                if i == 0:continue
                strline = fhA.next()
                A.append(strline.strip('\n').split('\t'))
            for i,fhB in enumerate(self.__ssnroisB):
                strline = fhB.next()
                B.append(strline.strip('\n').split('\t'))
            mA = []
            mB = []
            #test if NA in values
            try:
                NA_found = 0;
                for i,ln in enumerate(A):
                    if ln[self.__chr_ndx] == '#chr':continue
                    if ('NA' in A[i][self.__sssnroi_ndx]):
                        NA_found = 1
                        break                        # quit processing of ssnroisA
                    else:    
                        mA.extend([ A[i][self.__sssnroi_ndx].split(',') ])              
                    if (not NA_found):          # only process ssnroisB if no NA has been foun in ssnroisA
                        for i,ln in enumerate(B):
                            if ('NA' in B[i][self.__sssnroi_ndx]):
                                NA_found = 1
                                break                    # quit processing of ssnroisB
                            else:    
                                mB.extend([ B[i][self.__sssnroi_ndx].split(',') ])
                if (NA_found != 0):                                
                    invalidComparisons  = invalidComparisons  +1
                    continue        
            except (IndexError,KeyError),err:
                print 'Missing line or sample in group object!\nProcessing line ',i,'\n Last samples: ',A,B
                continue

            # test if mA or mB meet pass threshold condition:
            meanA = numpy.mean( [ numpy.mean(map(float,samp)) for samp in mA ] )
            self.__testValA = self.__size_ssnroisA * self.__both_grps_min    
            meanB = numpy.mean( [ numpy.mean(map(float,samp)) for samp in mB ] )
            self.__testValB = self.__size_ssnroisB * self.__both_grps_min    
                
            if ( ( meanA  < self.__testValA) and   ( meanB  < self.__testValB) ): 
                invalidComparisons  = invalidComparisons  + 1     


        validComparisons = allComparisons - invalidComparisons
        self.__fdr_cnt = validComparisons

        #output loci numbers
        print 'Number of loci in feature: ',str(allComparisons)
        print 'Number of invalid loci: ', str(invalidComparisons)
        print 'Number of loci used for FDR: ', str(self.__fdr_cnt)       


    #####
    ###
    ##
    #    Read line and run GEE
    ##
    ###
    #####
        # set statistic objects empty or = 0
        # nzfrac_list is a list of the non-zero fraction for each tested loci
        nzfrac_list = []
        # sterr_list is a list of the average standard error of each tested loci
        samp_sterr_list = []
        avg_sterr_list = []
        sterr_list = []
        #make DMR lists
        dmr_nzfrac = []
        dmr_samp_sterr = []
        dmr_avg_sterr = []
        dmr_sterr = []

        # seek to begining of iterators
        for file in self.__itlistA:
            file.seek(0)
        for file in self.__itlistB:
            file.seek(0)

        print 'Starting iteration for GEE fit'
        #use first file in group A to iterate over lines
        #zero line counter
        linecnt = 0
        print self.__itlistA[0]
        for itline in self.__itlistA[0]:
            #populate line lists
            A = []
            B = []
            #populate first item in A with list from iterator
            A.append(itline.strip('\n').split('\t'))
            #check for header line and skip
            if str(A[0][0][0]) == '#':
                #iterate through header in each file
                for fhA in self.__itlistA[1:]:
                    if str(fhA) == str(self.__itlistA[0]): continue
                    A.append(fhA.next().strip('\n').split('\t'))
                for fhB in self.__itlistB:
                    B.append(fhB.next().strip('\n').split('\t'))
                #skip to next line
                continue
            #populate A and B with list from each file
            for fhA in self.__itlistA[1:]:
                #skip file handle used as iterator
                if str(fhA) == str(self.__itlistA[0]): continue
                A.append(fhA.next().strip('\n').split('\t'))
            for fhB in self.__itlistB:
                B.append(fhB.next().strip('\n').split('\t'))

        
            ###
            # process list of lines
            ###
            
	    #make empty position and value vectors
            mA = []
            mB = []
            pA = []
            pB = []
            try:
                NA_found = 0;
                #iterate over samples in A:
                #   list of lists
                #   each list in "A" is the current line for each sample in groupA
                for i,ln in enumerate(A):
                    if ln[self.__chr_ndx] == '#chr':continue
                    if ('NA' in A[i][self.__sssnroi_ndx]):
                        NA_found=1
                    else:    
                        mA.extend([ A[i][self.__sssnroi_ndx].split(',') ])              
                        pA.extend([ A[i][self.__ssnroi_pos_ndx].split(',') ])

                for i,ln in enumerate(B):
                    if ('NA' in B[i][self.__sssnroi_ndx]):
                        NA_found=1
                    else:    
                        mB.extend([ B[i][self.__sssnroi_ndx].split(',') ])
                        pB.extend([ B[i][self.__ssnroi_pos_ndx].split(',') ])

                if (NA_found == 0):    
                    org_line = '\t'.join( [ A[0][x] for x in sorted( range( len(ln) ) ) if x != self.__sssnroi_ndx and x != self.__ssnroi_pos_ndx ])
                    org_line = org_line.strip('\n')         
                else:
                    NA_line = '\t'.join( [ A[0][x] for x in sorted( range( len(ln) ) ) if x != self.__sssnroi_ndx and x != self.__ssnroi_pos_ndx ])
                    NA_line= NA_line.strip('\n') 

                    self.__out.write(NA_line)    
                    self.__out.write('\tNAN\tNAN\tNAN\tNAN\n')
                    
                    continue

            except (IndexError,KeyError),err:
                exp_line = '\t'.join( [ A[0][x] for x in sorted( range( len(ln) ) ) if x != self.__sssnroi_ndx ]) 
                exp_line = exp_line.strip('\n') 
                
                self.__out.write(exp_line)            
                self.__out.write('\tNdxErr\tNdxErr\tNdxErr\tNdxErr\n')
                
                continue

            
            #generate loci data and info
            cgpos = A[0][self.__ssnroi_pos_ndx].split(',')
            print 'Length of cgpos: ',len(cgpos)
            print 'Length of values in A: '
            for i in mA:
                print len(i)
            print 'Length of values in B: '
            for i in mB:
                print len(i)                
            ch = A[0][self.__chr_ndx]
            start = A[0][self.__start_ndx]
            stop = A[0][self.__stop_ndx]
            loc_id = A[0][self.__loc_name_ndx]
            sumA = numpy.sum( [ numpy.mean(map(float,samp)) for samp in mA ] )
            avgA = numpy.mean(sumA)
            testValA = self.__size_ssnroisA * self.__both_grps_min    
            testValB = self.__size_ssnroisB * self.__both_grps_min    
            sumB = numpy.sum( [ numpy.mean(map(float,samp)) for samp in mB ] )
            avgB = numpy.mean(sumB)
            #   Determing the fraction of non-zero values
            #    to make a plot of standard err vs nonzero fraction, append vals to ste_list and nzfac_list
            nzcnt = 0
            poscnt = 0
            Aval = [ [] for i in mA[0] ]
            Bval = [ [] for i in mB[0] ]
            for s,samp in enumerate(mA):
                for i,val in enumerate(samp):
                    Aval[i].append(val)
                    poscnt += 1
                    if float(val) != 0.0: nzcnt += 1
            
            for s,samp in enumerate(mB):
                for i,val in enumerate(samp):
                    Bval[i].append(val)
                    poscnt += 1
                    if float(val) != 0.0: nzcnt += 1
            # compute non-zero fraction 
            nzfrac = float(nzcnt)/float(poscnt)
            nzfrac_list.append(nzfrac)
          
            # standard error of each CpG averaged over the feature
            avg_sterr = [ numpy.std(map(float,Aval[i]) + map(float,Bval[i])) / numpy.sqrt(len(Aval[i] + Bval[i])) for i,al in enumerate(Aval) ]
            avgerr = numpy.mean(avg_sterr)
            avg_sterr_list.append(avgerr)
            #standard error of entire feature
            values = []
            for i,al in enumerate(Aval):
                for j,v in enumerate(Aval[i]):
                    values.append(v)
                for j,v in enumerate(Bval[i]):
                    values.append(v)
            ster = numpy.std(map(float,values))/numpy.sqrt(len(map(float,values)))
            sterr_list.append(ster)
            #standard error of each sample, averaged
            samp_err = []
            for i,l in enumerate(mA):
                samp_err.append(numpy.std(map(float,l))/numpy.sqrt(len(l)))
            for i,l in enumerate(mB):
                samp_err.append(numpy.std(map(float,l))/numpy.sqrt(len(l)))
            samp_sterr_list.append(numpy.mean(samp_err))

            # test if mA or mB meet pass threshold condition:
            if ( ( float(sumA)  >= float(testValA)) or  ( float(sumB)  >= float(testValB) ) ) : 

                #####
                ###
                #  GEE function with logit link test 
                ###
                #####
                #used to only include number of CpGs in GEE
                (pval, fval, llval) = self.__GEE(mA,mB, self.__fdr_cnt)

                spval=str(pval)

                #test for class 1, 2 or 3 DMRs
                    #class 1: fdr < 0.05, nzfrac > 0.5
                    #class 2: fdr < 0.05, logit link < 0.05, any nzfrac
                    #class 3: fdr < 0.05, any 0.05, any nzfrac 
                #class 1
                if fval <= 0.05 and nzfrac > 0.5:
                    DMRclass = 1
                    DMRid = 'class_1'
                #class 2
                elif fval <= 0.05 and llval <= 0.05 and nzfrac >= 0.5:
                    DMRclass = 2
                    DMRid = 'class_2'
                #class 3
                elif fval <= 0.05:
                    DMRclass = 3
                    DMRid = 'class_3'
                #not DMR
                else:
                    DMRclass = 0
                    DMRid = ''    

                if DMRclass != 0:
                    print "pval,fval,class: ", spval,fval,DMRclass

                spval=str(pval)
                
                
                print "pval,fval,class: ", spval,fval,DMRclass
                line = org_line + '\t'+ str(pval) + '\t' + str(fval) + '\t'+ str(DMRclass) + '\t' + str(nzfrac) + '\t' + str(avgA)+ '\t' + str(avgB)   
                
                self.__out.write(line + '\n')      
                     
                # write to fltrd_out only if fval<=.05
                if ((fval != None ) and (fval<=.05 and fval >= 0)):       
                    plotout = str(self.__plots_out)+'/DMR_'+str(loc_id)+'_'+str(ch)+'_'+str(start)+'-'+str(stop)+str(round(fval,3))+'_plot.png'
                    title = str(loc_id)+' '+str(ch)+':'+str(start)+'-'+str(stop)+' FDR pval: '+str(fval)

                    self.__fltrd_out.write(line+'\n')

                   
                    
                else:
                    try:
                        fval = float(fval)
                        #plot nonDMR loci            
                        plotout = str(self.__plots_out)+'/'+str(loc_id)+'_'+str(ch)+'_'+str(start)+'-'+str(stop)+'_qv_'+str(round(fval,3))+'_plot.png'
                        title = str(loc_id)+' '+str(ch)+':'+str(start)+'-'+str(stop)+' FDR pval: '+str(fval)
 
                    except(UnboundLocalError,TypeError):
                        print 'fval failed converstion to float ',fval
                        pass                   
            else : 
                self.__out.write(org_line + '\t'+ 'dataFailedTest' + '\t' + 'NA' + '\t' + '\t' +'NA' +str(nzfrac)+'\t'+ str(avgA)+ '\t' + str(avgB) + '\n')    
                    

            #increment line counter
            linecnt += 1
            if (linecnt % 1000) == 0:print '\tprocessed loci number: ',linecnt


        # plot std err vs non-zero fraction
        plt.figure()
        plt.plot(nzfrac_list,sterr_list,'ko')
        plt.title('Standard Error of all loci ('+str(len(nzfrac_list))+' loci )')
	plt.savefig(plots_out+'_nzfrac_vs_stdERR.png',format='png')
        plt.close()
        # avg std err
        plt.figure()
        plt.plot(nzfrac_list,avg_sterr_list,'ko')
        plt.title('Average Standard Error of each CpG in all loci ('+str(len(nzfrac_list))+' loci )')
        
	plt.savefig(plots_out+'_nzfrac_vs_avg_stdERR.png',format='png')
        plt.close()    
        #sample std err
        plt.figure()
        plt.plot(nzfrac_list,samp_sterr_list,'ko')
        plt.title('Average Standard Error of each sample in all loci ('+str(len(nzfrac_list))+' loci )')
        plt.savefig(plots_out+'_nzfrac_vs_sample_stdERR.png',format='png')
        plt.close()   


        # plot std err vs non-zero fraction
        plt.figure()
        plt.plot(dmr_nzfrac,dmr_sterr,'ko')
        plt.title('Standard Error of all DMR loci ('+str(len(dmr_sterr))+' loci )')
        plt.savefig(plots_out+'_DMR_nzfrac_vs_stdERR.png',format='png')
        plt.close()
        # avg std err
        plt.figure()
        plt.plot(dmr_nzfrac,dmr_avg_sterr,'ko')
        plt.title('Average Standard Error of each CpG in all DMR loci ('+str(len(dmr_sterr))+' loci )')
        plt.savefig(plots_out+'_DMR_nzfrac_vs_avg_stdERR.png',format='png')
        plt.close()    
        #sample std err
        plt.figure()
        plt.plot(dmr_nzfrac,dmr_samp_sterr,'ko')
        plt.title('Average Standard Error of each sample in all DMR loci ('+str(len(dmr_sterr))+' loci )')
        plt.savefig(plots_out+'_DMR_nzfrac_vs_sample_stdERR.png',format='png')

        plt.close()  


        return

