#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-


###
#COPYRIGHT © 2013 THE OHIO STATE UNIVERSITY
#ALL RIGHTS RESERVED
#
#PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
#REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
#EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
#LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
#DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
#THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
#PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
#SPECIFIC, WRITTEN PRIOR AUTHORIZATION.
#
#THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
#STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
#WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
#IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
#MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
#UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
#INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
#CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
#EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
#DAMAGES.
###

try:    
    import os
    import sys
    import gzip
    import cPickle as pickle
    import argparse
    import datetime
    import csv
    
except (ImportError):
    print 'ERROR:\n\tOne or more modules not able to be imported.\n\tCheck that proper modules are available\n\tRequired modules:\n\t\tos\n\t\tsys\n\t\tgzip\n\t\tpickle\n\t\targparse\n\t\tdatetime\n\t\tcsv'
    exit()

def mmroi(bed_name, cgb_name, out_name):
   
    # open the output file        
    try: out = open(out_name ,'w')
    except (IOError):
        print 'ERROR:\n\tOutput file cannot be opened for writing!\n\t\"',out_name,'\": No such file!\n\t Exiting'
    	return

    # open cgb file   
    try:
        cgbhndl = open(cgb_name,'r')
    except (IOError),err:
        print 'ERROR:\n\tInput .cgb file cannot be opened for reading!\n\t\"',cgb_name,'\": Failed to read file!\n\t Exiting'
    
    # load cgb into memory cgb[<chr>][<CpG_position>] = [<coverage>,<coverage_per_million>]
    try:
        cgb = pickle.load(cgbhndl)	 
    except (EOFError,ValueError,TypeError),err:
        print 'ERROR:\n\tInput .cgb file cannot be loaded to memory!\n\t\"',out_name,'\": Cannot read file!\n\t Exiting'
        return
    # open the bed file with csv reader    
    try: bedhndl = open(bed_name,'r')
    except (IOError):
        print 'ERROR:\n\tInput BED file cannot be opened for reading!\n\t\"',bed_name,'\": No such file!\n\t Exiting'
        return
    try: bed = csv.reader(bedhndl,delimiter='\t')
    except (IOError):
        print 'ERROR:\n\tInput BED file cannot be opened for reading!\n\t\"',bed_name,'\": No such file!\n\t Exiting'
        return

    #check for header in BED
    header = bed.next()
    if header[0][0] == '#':
        #remove '#'
        newheader = []
        for obj in header:
            newheader.append(obj.strip('#'))
        orig_header = header
        header = newheader
        #check for required fields
        hcheck = True
        for chk in ['chr','start','stop','name']:
            try: 
                header.index(chk)
            except(IndexError):
                hcheck = False
                print 'ERROR:\n\tMissing required header: ',chk
        if hcheck: return
    else:
        header = ['chr','start','stop','name']
        print 'WARNING: No header line in BED file (denoted as header by leading "#").\nAssuming standard BED format:\nFields:\n1) chr\n2) start\n3) stop\n4) name'
    

    # create an list that stores the output order of columns
    outorder = []
    stdout = ['chr','start','stop','name']
    outorder.extend(stdout)
    outorder.extend( [ col for col in header if col not in stdout ] ) 

    # output header line to mmroi    
    out.write('#'+'\t'.join(outorder)+'\tmmroi\tpos\n')        
    
    # find CpGs in cgb that fall within each region in the BED file
    missing_chr = []
    for row in bed:
        if row[0][0] == '#' or not row: continue
        pnts = []
        locs = []
        #set BED values
        ch = row[header.index('chr')]
        start = int(row[header.index('start')])
        stop = int(row[header.index('stop')])
        # add values from cgb
        if cgb[ch]:
            for i in xrange(start,stop):
                if i in cgb[ch].keys():
                    try:
                        pnts.append(float(cgb[ch][i]))
                        cp=int(i)-int(start)
                        locs.append(int(cp))
                    except (KeyError):
                        continue
                else: continue

        else: missing_chr.append(ch)
        
        # create output line    
        outline = [ row[header.index(fld)] for fld in outorder ]
        if len(pnts) == 0: 
            outline.extend(['NA','NA'])
        else:
            outline.extend([str(','.join(map(str,pnts))),str(','.join(map(str,locs)))])
        # output line to file
        out.write('\t'.join(map(str,outline))+'\n')

    # output any missing chr from cgb
    if missing_chr:
        print 'WARNING:\nMissing chromosomes in cgb file!\nChromosomes included in BED file not able to be found in cgb file:', '\n\t'.join(missing_chr)

    
def main(arg1, arg2, arg3):
  
    import os
    bed_name = arg1
    cgb_name = arg2
    output_name = arg3
    #print "bed_name = ",arg1
    #print "cgb_name = ",arg2
    #print "output_name =",arg3
    mmroi(bed_name, cgb_name, output_name)
   
            
    return
           
           
  

if __name__ == '__main__':

    import argparse
    import sys
    parser = argparse.ArgumentParser(description='Generate mmroi files')
    parser.add_argument('--out',dest='out',nargs=1,type=str,
                        default=[sys.stdout],help='Output mmroi-output file name.')
    parser.add_argument('--cgb',dest='cgb',nargs='+',type=str,
                        default=[],help='input - cgb sample file name .')
    parser.add_argument('--bed',dest='bed',nargs='+',type=str,
                        default=[],help='input - bed feature file name.')
    args = parser.parse_args()
    main(args.bed[0],args.cgb[0],args.out[0])
